/* jruby.c
   provides a ruby interface for Jed
   Copyright (C) 1997 MAEDA Shugo */

#include <stdio.h>

#include "config.h"
#include "jed-feat.h"
#include "jruby.h"
#include "node.h"

#include "buffer.h"
#include "keymap.h"
#include "file.h"
#include "ins.h"
#include "ledit.h"
#include "screen.h"
#include "window.h"
#include "display.h"
#include "search.h"
#include "misc.h"
#include "replace.h"
#include "paste.h"
#include "sysdep.h"
#include "cmds.h"
#include "text.h"
#include "hooks.h"
#include "undo.h"
#include "kanji.h"

extern VALUE ruby_top_self;

VALUE mJed;
VALUE rb_hook_list;

static char *RUBY_PROMPT = "ruby>";
static char *RUBY_ERR_BUFFER = "*ruby backtrace*";
extern int (*complete_open)(char *);

/* copied from eval.c */
#define TAG_RETURN	0x1
#define TAG_BREAK	0x2
#define TAG_NEXT	0x3
#define TAG_RETRY	0x4
#define TAG_REDO	0x5
#define TAG_RAISE	0x6
#define TAG_THROW	0x7
#define TAG_FATAL	0x8
#define TAG_MASK	0xf

static void error_pos()
{
    char buf[BUF_SIZE];
    ID last_func = rb_frame_last_func();

    if (ruby_sourcefile) {
	if (last_func) {
	    snprintf(buf, BUF_SIZE, "%s:%d:in `%s'",
		     ruby_sourcefile, ruby_sourceline,
		     rb_id2name(last_func));
	}
	else {
	    snprintf(buf, BUF_SIZE, "%s:%d", ruby_sourcefile, ruby_sourceline);
	}
	insert_string(buf);
    }
}

void ruby_fail(int state)
{
    extern VALUE ruby_errinfo;
    VALUE einfo, errat;
    char *errmsg;
    char buf[BUF_SIZE];

    switch (state) {
      case TAG_RETURN:
	error_pos();
	insert_string(": unexpected return\n");
	return;
      case TAG_NEXT:
	error_pos();
	insert_string(": unexpected next\n");
	return;
      case TAG_BREAK:
	error_pos();
	insert_string(": unexpected break\n");
	return;
      case TAG_REDO:
	error_pos();
	insert_string(": unexpected redo\n");
	return;
      case TAG_RETRY:
	error_pos();
	insert_string(": retry outside of rescue clause\n");
	return;
      case TAG_RAISE:
      case TAG_FATAL:
        break;
      default:
	rb_bug("Unknown longjmp status %d", state);
	return;
    }

    einfo = rb_obj_as_string(ruby_errinfo);
    if (NIL_P(ruby_errinfo))
	errat = Qnil;
    else
	errat = rb_funcall(ruby_errinfo, rb_intern("backtrace"), 0);

    set_buffer(RUBY_ERR_BUFFER);
    bob();

    if (RSTRING(einfo)->len == 0) {
	errmsg = "unhandled exception";
	insert_string("unhandled exception.\n");
    } else {
	errmsg = RSTRING(einfo)->ptr;
	insert_string(RSTRING(einfo)->ptr);
	if (RSTRING(einfo)->ptr[RSTRING(einfo)->len - 1] != '\n') {
	    insert_string("\n");
	}
    }

    if (!NIL_P(errat)) {
	int i;
	struct RArray *ep = RARRAY(errat);

#define TRACE_MAX (TRACE_HEAD+TRACE_TAIL+5)
#define TRACE_HEAD 8
#define TRACE_TAIL 5

	for (i = 1; i < ep->len; i++) {
	    sprintf(buf, "\tfrom %s\n", RSTRING(ep->ptr[i])->ptr);
	    insert_string(buf);
	    if (i == TRACE_HEAD && ep->len > TRACE_MAX) {
		sprintf(buf, "\t ... %d levels...\n",
			ep->len - TRACE_HEAD - TRACE_TAIL);
		insert_string(buf);
		i = ep->len - TRACE_TAIL;
	    }
	}
    }

    msg_error(errmsg);
}

void ruby_eval(char *str)
{
    VALUE ret;
    int state;

    ret = rb_eval_string_protect(str, &state);
    if (state)
	ruby_fail(state);
    else
	message(RSTRING(rb_inspect(ret))->ptr);
}

int is_ruby_command(char *name)
{
    ID id;

    if (name == NULL || strlen(name) == 0) return Qfalse;
    id = rb_intern(name);
    return rb_method_boundp(rb_cObject, id, NOEX_PUBLIC);
}

static VALUE ruby_call_method(ID *id)
{
    return rb_funcall(ruby_top_self, *id, 0);
}

void ruby_execute_command(char *name)
{
    int state;
    ID id;
    
    id = rb_intern(name);
    if (rb_method_boundp(rb_cObject, id, NOEX_PUBLIC)) {
	rb_protect(ruby_call_method, (VALUE) &id, &state);
	if (state) ruby_fail(state);
    }
}

void ruby_load(char *fname)
{
    rb_load(rb_str_new2(fname), 0);
}

int ruby_evaluate_cmd()
{
    int n;
    char *tmp;
    
    if (MiniBuffer == NULL) complete_open = NULL;
    if ((tmp = (char *) read_from_minibuffer(RUBY_PROMPT, 0, NULL, &n))
	== NULL) return 1;
    SLang_Error = 0;
    Suspend_Screen_Update = 1;
    
    ruby_eval(tmp);
    SLFREE(tmp);
    
    SLang_Error = 0;
   
    return 0;
}

static VALUE mjed_evaluate_cmd(VALUE obj)
{
    ruby_evaluate_cmd();
    return Qnil;
}

static VALUE mjed_internal_function_p(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return RBOOL(SLang_find_key_function(RSTRING(str)->ptr,
					 CBuf->keymap) != NULL);
}

static VALUE mjed_call_internal_function(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    if (SLang_find_key_function(RSTRING(str)->ptr,
				CBuf->keymap) != NULL) {
	call_cmd(RSTRING(str)->ptr);
	return Qtrue;
    } else {
	return Qnil;
    }
}

static VALUE mjed_defined_in_slang_p(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return RBOOL(SLang_is_defined(RSTRING(str)->ptr));
}

static VALUE mjed_eval_slang_string(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    if (SLang_is_defined(RSTRING(str)->ptr)) {
	SLang_load_string(RSTRING(str)->ptr);
	return Qtrue;
    } else {
	return Qfalse;
    }
}

static VALUE mjed_exit_jed(VALUE obj)
{
    exit_jed();
    return Qnil;
}

static VALUE mjed_quit_jed(VALUE obj)
{
    quit_jed();
    return Qnil;
}

static VALUE mjed_set_status_line(VALUE obj,
				 VALUE format, VALUE local)
{
    int n;
    Check_Type(format, T_STRING);
    n = FIX2INT(local);
    set_status_format(RSTRING(format)->ptr, &n);
    return Qnil;
}

static VALUE mjed_update(VALUE obj, VALUE force)
{
    int f = RTEST(force);
    update_cmd(&f);
    return Qnil;
}

static VALUE mjed_message(int argc, VALUE *argv,
			  VALUE obj)
{
    VALUE str = rb_str_new2("");
    VALUE str2;
    int i;
    
    for (i = 0; i < argc; i++) {
	if (TYPE(argv[i]) == T_STRING)
	    str2 = argv[i];
	else
	    str2 = rb_obj_as_string(argv[i]);
	str = rb_str_plus(str, str2);
    }
    message(RSTRING(str)->ptr);

    return str;
}

static VALUE mjed_flush_message(int argc, VALUE *argv,
			  VALUE obj)
{
    VALUE str = rb_str_new2("");
    VALUE str2;
    int i;
    
    for (i = 0; i < argc; i++) {
	if (TYPE(argv[i]) == T_STRING)
	    str2 = argv[i];
	else
	    str2 = rb_obj_as_string(argv[i]);
	str = rb_str_plus(str, str2);
    }
    flush_message(RSTRING(str)->ptr);

    return str;
}

static VALUE mjed_show_error(int argc, VALUE *argv,
			     VALUE obj)
{
    VALUE str = rb_str_new2("");
    VALUE str2;
    int i;
    
    for (i = 0; i < argc; i++) {
	if (TYPE(argv[i]) == T_STRING)
	    str2 = argv[i];
	else
	    str2 = rb_obj_as_string(argv[i]);
	str = rb_str_plus(str, str2);
    }
    msg_error(RSTRING(str)->ptr);

    return str;
}

static VALUE mjed_read_from_minibuffer(VALUE obj,
				       VALUE prompt,
				       VALUE def,
				       VALUE stuff)
{
    char *buf;
    int n;
    VALUE result;

    Check_Type(prompt, T_STRING);
    Check_Type(def, T_STRING);
    Check_Type(stuff, T_STRING);
    complete_open = NULL;
    if ((buf = read_from_minibuffer(RSTRING(prompt)->ptr,
				    RSTRING(def)->ptr,
				    RSTRING(stuff)->ptr, &n)) == NULL) {
	result = rb_str_new2("");
    } else {
	result = rb_str_new2(buf);
	SLFREE(buf);
    }

    return result;
}

extern VALUE mjed_read_with_completion(VALUE obj, VALUE vprompt,
				       VALUE vdflt, VALUE vstuff,
				       VALUE vtype);

static VALUE mjed_beep(VALUE obj)
{
    tt_beep();
    return Qnil;
}

static VALUE mjed_input_pending(VALUE obj, VALUE tsec)
{
    int n;

    Check_Type(tsec, T_FIXNUM);
    n = FIX2INT(tsec);
    return RBOOL(input_pending(&n));
}

static VALUE mjed_send_string_to_terminal(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    if (Batch) {
	fputs(RSTRING(str)->ptr, stdout);
    } else tt_write_string(RSTRING(str)->ptr);
    return Qnil;
}

static VALUE mjed_pause(VALUE obj, VALUE ms)
{
    if (FIX2INT(ms) < 0) return Qnil;
    sys_pause(FIX2INT(ms));
}

static VALUE mjed_recenter(int argc, VALUE *argv,
			   VALUE obj)
{
    VALUE num;
    int n;

    if (rb_scan_args(argc, argv, "01", &num) == 0) {
	n = 0;
    } else {
	Check_Type(num, T_FIXNUM);
	n = FIX2INT(num);
    }
    recenter(&n);
    return Qnil;
}

static VALUE mjed_get_yes_no(VALUE obj, VALUE question)
{
    Check_Type(question, T_STRING);
    return RBOOL(get_yes_no(RSTRING(question)->ptr));
}

static VALUE mjed_find_file(int argc, VALUE *argv,
			    VALUE obj)
{
    VALUE filename;

    if (rb_scan_args(argc, argv, "01", &filename) == 0) {
	find_file();
    } else {
	Check_Type(filename, T_STRING);
	find_file_in_window(RSTRING(filename)->ptr);
    }
}

static VALUE mjed_create_keymap(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    create_keymap(RSTRING(name)->ptr);

    return Qnil;
}

static VALUE mjed_copy_keymap(VALUE obj, VALUE newname, VALUE basename)
{
    Check_Type(newname, T_STRING);
    Check_Type(basename, T_STRING);
    jed_copy_keymap(RSTRING(newname)->ptr, RSTRING(basename)->ptr);

    return Qnil;
}

static VALUE mjed_define_key(int argc, VALUE *argv,
			     VALUE obj)
{
    VALUE method, key, keymap;
    char *name;
    SLKeyMap_List_Type *kmap;

    if (rb_scan_args(argc, argv, "21", &method, &key, &keymap) == 2) {
	kmap = Global_Map;
    } else {
	Check_Type(keymap, T_STRING);
	if ((kmap = SLang_find_keymap(RSTRING(keymap)->ptr)) == NULL)
	  rb_raise(rb_eRuntimeError, "unknown keymap %s", RSTRING(keymap)->ptr);
    }
    switch (TYPE(method)) {
    case T_FIXNUM:
	name = rb_id2name((ID) FIX2INT(method));
	break;
    case T_STRING:
	name = RSTRING(method)->ptr;
	break;
    default:
	rb_raise(rb_eTypeError, "wrong argument type %s (expected String or Fixnum)",
		   TYPE(method));
    }

    Check_Type(key, T_STRING);
    SLang_define_key(RSTRING(key)->ptr, name, kmap);

    return Qnil;
}

static VALUE mjed_undefine_key(int argc, VALUE *argv,
			       VALUE obj)
{
    VALUE key, keymap;
    SLKeyMap_List_Type *kmap;
   
    if (rb_scan_args(argc, argv, "11", &key, &keymap) == 1) {
	kmap = Global_Map;
    } else {
	Check_Type(keymap, T_STRING);
	if ((kmap = SLang_find_keymap(RSTRING(keymap)->ptr)) == NULL)
	  rb_raise(rb_eRuntimeError, "unknown keymap %s", RSTRING(keymap)->ptr);
    }
    Check_Type(key, T_STRING);
    SLang_undefine_key(RSTRING(key)->ptr, Global_Map);

    return Qnil;
}

static VALUE mjed_getkey(VALUE obj)
{
    return INT2FIX(jed_getkey());
}

static VALUE mjed_ungetkey(VALUE obj, VALUE ch)
{
    int n;

    Check_Type(ch, T_FIXNUM);
    n = FIX2INT(ch);
    ungetkey(&n);
    return Qnil;
}

static VALUE mjed_add_completion(VALUE obj, VALUE method)
{
    char *name;

    switch (TYPE(method)) {
    case T_FIXNUM:
	name = rb_id2name(FIX2UINT(method));
	break;
    case T_STRING:
	name = RSTRING(method)->ptr;
	break;
    default:
	rb_raise(rb_eTypeError, "wrong argument type %s (expected String or Fixnum)",
		   TYPE(method));
    }
    add_to_completion(name);
    return rb_str_new2(name);
}

static VALUE mjed_set_color(VALUE obj, VALUE tgt,
			   VALUE fg, VALUE bg)
{
    Check_Type(tgt, T_STRING);
    Check_Type(fg, T_STRING);
    Check_Type(bg, T_STRING);
    set_term_colors(RSTRING(tgt)->ptr,
		    RSTRING(fg)->ptr,
		    RSTRING(bg)->ptr);
    return Qnil;
}

static VALUE mjed_set_mode(VALUE obj, VALUE mode, VALUE flags)
{
    Check_Type(mode, T_STRING);
    Check_Type(flags, T_FIXNUM);

/* free ha? */
   CBuf->mode_string = RSTRING(mode)->ptr;
   CBuf->modes = FIX2INT(flags);
#if JED_HAS_LINE_ATTRIBUTES
   CBuf->min_unparsed_line_num = 1;
   CBuf->max_unparsed_line_num = Max_LineNum + CBuf->nup;
#endif
    return Qnil;
}

static VALUE mjed_keymap_p(VALUE obj, VALUE keymap)
{
    Check_Type(keymap, T_STRING);
    if (SLang_find_keymap(RSTRING(keymap)->ptr))
      return Qtrue;
    else
      return Qfalse;
}

static VALUE mjed_what_keymap(VALUE obj)
{
    return rb_str_new2(what_keymap());
}

static VALUE mjed_use_keymap(VALUE obj, VALUE keymap)
{
    Check_Type(keymap, T_STRING);
    if (SLang_find_keymap(RSTRING(keymap)->ptr) == NULL)
      rb_raise(rb_eRuntimeError, "unknown keymap %s", RSTRING(keymap)->ptr);
    use_keymap(RSTRING(keymap)->ptr);
    return Qnil;
}

static VALUE mjed_dump_key_bindings(VALUE obj, VALUE keymap)
{
    Check_Type(keymap, T_STRING);
    dump_bindings(RSTRING(keymap)->ptr);
    return Qnil;
}

static VALUE mjed_create_syntax_table(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    create_syntax_table(RSTRING(name)->ptr);
    return Qnil;
}

static VALUE mjed_define_syntax(int argc, VALUE *argv,
				VALUE obj)
{
    VALUE what0, name0, val1, val2;
    char *name;
    int what;
    Syntax_Table_Type *table;
    int c2;
    unsigned int i;
    char *s1 = NULL, *s2 = NULL;
    unsigned char lut[256], *s;

    if (rb_scan_args(argc, argv, "31",
		     &name0, &what0, &val1, &val2) > 2) {
	Check_Type(name0, T_STRING);
	name = RSTRING(name0)->ptr;
	Check_Type(what0, T_FIXNUM);
	what = FIX2INT(what0);
    }
    
    table = find_syntax_table(name, 1);
    if (table == NULL)
      rb_raise(rb_eRuntimeError, "no such table `%s'", name);
   
    switch (what) {
    case '%':
	Check_Type(val1, T_STRING);
	s1 = SLang_create_slstring(RSTRING(val1)->ptr);
	if (argc == 4)
	  {
	     Check_Type(val2, T_STRING);
	     s2 = SLang_create_slstring(RSTRING(val2)->ptr);
	  }
	
	table->char_syntax[(unsigned char) *s1] |= COMMENT_SYNTAX;

	if ((s2 == NULL) || (*s2 == 0) || (*s2 == '\n'))
	  {
	     SLang_free_slstring (table->eol_comment_start);
	     table->eol_comment_start = s1;
	     table->eol_comment_len = strlen (s1);
	     SLang_free_slstring (s2);
	     break;
	  }

	table->char_syntax[(unsigned char) *s2] |= COMMENT_SYNTAX;

	SLang_free_slstring (table->comment_start);
	SLang_free_slstring (table->comment_stop);
	table->comment_start = s1;
	table->comment_stop = s2;
	table->comment_start_len = strlen (s1);
	table->comment_stop_len = strlen (s2);
	break;
	
    case '\\':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_FIXNUM);
	    c2 = FIX2INT(val1);
	}
	table->char_syntax[(unsigned char) c2] |= QUOTE_SYNTAX;
	table->quote_char = (unsigned char) c2;
	break;
	
    case '#':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_FIXNUM);
	    c2 = FIX2INT(val1);
	}
	table->preprocess = (unsigned char) c2;
	break;
      
    case '\'':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_FIXNUM);
	    c2 = FIX2INT(val1);
	}
	table->char_syntax[(unsigned char) c2] |= STRING_SYNTAX;
	table->char_char = (unsigned char) c2;
	break;
    case '"':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_FIXNUM);
	    c2 = FIX2INT(val1);
	}
	table->char_syntax[(unsigned char) c2] |= STRING_SYNTAX;
	table->string_char = (unsigned char) c2;
	break;
	
    case '<':
    case '>':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	}
	s2 = s1;
	while (*s2 != 0) {
	     if (*(s2 + 1) == 0) break;
	     table->char_syntax[(unsigned char) *s2] |= HTML_START_SYNTAX;
	     table->char_syntax[(unsigned char) *(s2 + 1)] |= HTML_END_SYNTAX;
	     s2 += 2;
	}
	s2 = NULL;
	break;
	
    case '(':
    case ')':
	if (rb_scan_args(argc, argv, "40",
			 &name0, &what0, &val1, &val2) == 4) {
	    Check_Type(val1, T_STRING);
	    Check_Type(val2, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	    s2 = RSTRING(val2)->ptr;
	}
	
	i = strlen (s1);
	if (i != strlen (s2)) {
	    msg_error ("Delimiter set does not match.");
	}
	while (i > 0) {
	    unsigned char ch1, ch2;
	    i--;
	    ch1 = (unsigned char) s1[i]; ch2 = (unsigned char) s2[i];
	    table->char_syntax[ch1] |= OPEN_DELIM_SYNTAX;
	    table->char_syntax[ch2] |= CLOSE_DELIM_SYNTAX;
	    table->matching_delim[ch2] = ch1;
	    table->matching_delim[ch1] = ch2;
	}
	break;
	
    case '+':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	}
	for (i = 0; i < 256; i++) table->char_syntax[i] &= ~OP_SYNTAX;

	s = (unsigned char *) s1;
	while (*s) {
	    table->char_syntax[*s] |= OP_SYNTAX;
	    s++;
	}
	break;
	
    case '0':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	}
	SLmake_lut (lut, (unsigned char *) s1, 0);

	for (i = 0; i < 256; i++) {
	    if (lut[i]) table->char_syntax[i] |= NUMBER_SYNTAX;
	    else table->char_syntax[i] &= ~NUMBER_SYNTAX;
	}
	break;
	
    case ',':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	}
	s = (unsigned char *) s1;
	for (i = 0; i < 256; i++) table->char_syntax[i] &= ~DELIM_SYNTAX;
	while (*s) {
	     table->char_syntax[*s] |= DELIM_SYNTAX;
	     s++;
	}
	break;
	
    case 'w':
	if (rb_scan_args(argc, argv, "30",
			 &name0, &what0, &val1) == 3) {
	    Check_Type(val1, T_STRING);
	    s1 = RSTRING(val1)->ptr;
	}
	SLmake_lut (lut, (unsigned char *) s1, 0);

	for (i = 0; i < 256; i++) {
	    if (lut[i]) table->char_syntax[i] |= WORD_SYNTAX;
	    else table->char_syntax[i] &= ~WORD_SYNTAX;
	}
	break;

    default:
	msg_error ("Bad parameter to define_syntax");
    }
    
    return Qtrue;
}

static VALUE mjed_use_syntax_table(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    if (find_syntax_table(RSTRING(name)->ptr, 1))
      use_syntax_table(RSTRING(name)->ptr);
    else
      rb_raise(rb_eRuntimeError, "no such table `%s'", RSTRING(name)->ptr);
    return Qnil;
}

#if JED_HAS_DFA_SYNTAX
static VALUE mjed_define_highlight_rule(VALUE obj, VALUE rule,
					VALUE colour, VALUE name)
{
    Check_Type(rule, T_STRING);
    Check_Type(colour, T_STRING);
    Check_Type(name, T_STRING);
    define_highlight_rule(RSTRING(rule)->ptr, RSTRING(colour)->ptr,
			  RSTRING(name)->ptr);
    return Qnil;
}

static VALUE mjed_build_highlight_table(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    build_highlight_table(RSTRING(name)->ptr);
    return Qnil;
}

static VALUE mjed_enable_highlight_cache(VALUE obj,
					 VALUE file, VALUE name)
{
    Check_Type(file, T_STRING);
    Check_Type(name, T_STRING);
    enable_highlight_cache(RSTRING(file)->ptr, RSTRING(name)->ptr);
    return Qnil;
}
#endif /* JED_HAS_DFA_SYNTAX */

static VALUE mjed_define_keywords(VALUE obj,
				  VALUE name, VALUE kwords,
				  VALUE len, VALUE tbl_num)
{
    int m, n;
      
    m = FIX2INT(len);
    n = FIX2INT(tbl_num);
    Check_Type(name, T_STRING);
    Check_Type(kwords, T_STRING);
    Check_Type(len, T_FIXNUM);
    Check_Type(tbl_num, T_FIXNUM);
    define_keywords(RSTRING(name)->ptr,
		    RSTRING(kwords)->ptr,
		    &m, &n);
    
    return Qnil;
}

static VALUE mjed_add_hook(int argc, VALUE *argv,
			   VALUE obj)
{
    VALUE name, proc, hooks;
    
    if (rb_scan_args(argc, argv, "11", &name, &proc) == 1) {
	proc = rb_f_lambda();
    }
    switch (TYPE(name)) {
    case T_FIXNUM:
	name = rb_str_new2(rb_id2name(FIX2INT(name)));
	break;
    case T_STRING:
	break;
    default:
	rb_raise(rb_eTypeError, 
	    "wrong argument type %s (expected String or Fixnum)",
	    TYPE(proc));
    }
    if (!rb_obj_is_kind_of(proc, rb_cProc)) {
	switch (TYPE(proc)) {
	case T_FIXNUM:
	case T_STRING:
	    break;
	default:
	    rb_raise(rb_eTypeError, 
		"wrong argument type %s (expected String or Fixnum or Proc)",
		TYPE(proc));
	}
    }
    if (TYPE(hooks = rb_hash_aref(rb_hook_list, name)) != T_ARRAY) {
	hooks = rb_ary_new();
	rb_hash_aset(rb_hook_list, name, hooks);
    }
    rb_ary_push(hooks, proc);
    return Qnil;
}

static VALUE mjed_run_hooks(int argc, VALUE *argv,
			    VALUE obj)
{
    VALUE name;
    VALUE dummy1, dummy2, dummy3, dummy4, dummy5,
	dummy6, dummy7, dummy8, dummy9;
    VALUE hooks, proc;
    ID id_call = rb_intern("call");
    
    rb_scan_args(argc, argv, "19", &name,
		 &dummy1, &dummy2, &dummy3, &dummy4, &dummy5,
		 &dummy6, &dummy7, &dummy8, &dummy9);
    argv++;
    argc--;
    switch (TYPE(name)) {
    case T_FIXNUM:
	name = rb_str_new2(rb_id2name(FIX2INT(name)));
	break;
    case T_STRING:
	break;
    default:
	rb_raise(rb_eTypeError, 
	    "wrong argument type %s (expected String or Fixnum)",
	    TYPE(proc));
    }
    if (TYPE(hooks = rb_hash_aref(rb_hook_list, name)) == T_ARRAY) {
	int i;
	
	for (i=0; i < RARRAY(hooks)->len; i++) {
	    proc = RARRAY(hooks)->ptr[i];
	    if (rb_obj_is_kind_of(proc, rb_cProc)) {
		rb_funcall2(proc, id_call, argc, argv);
	    } else {
		ID id;
		
		switch (TYPE(proc)) {
		case T_FIXNUM:
		    id = FIX2INT(proc);
		    break;
		case T_STRING:
		    id = rb_intern(RSTRING(proc)->ptr);
		    break;
		default:
		    rb_raise(rb_eRuntimeError, "Invalid hook");
		}
		if (rb_method_boundp(rb_cObject, id, NOEX_PUBLIC))
		    rb_funcall2(ruby_top_self, id, argc, argv);
	    }
	}
    }
    return Qnil;
}

struct run_hooks_arg {
    int count;
    VALUE *vector;
};

static VALUE run_hooks(struct run_hooks_arg *arg)
{
    mjed_run_hooks(arg->count, arg->vector, ruby_top_self);
    return Qnil;
}

void rb_run_hooks(int argc, VALUE *argv)
{
    int state;
    struct run_hooks_arg arg;

    arg.count = argc;
    arg.vector = argv;
    rb_protect(run_hooks, (VALUE) &arg, &state);
    if (state) ruby_fail(state);
}

#if !defined(pc_system) || defined(__GO32__)
static VALUE use_ansi_colors_getter()
{
    return RBOOL(*tt_Use_Ansi_Colors);
}

static void use_ansi_colors_setter(VALUE val)
{
    *tt_Use_Ansi_Colors = val;
}
#endif /* !defined(pc_system) || defined(__GO32__) */

#ifdef REAL_UNIX_SYSTEM
static VALUE backup_by_copy_getter()
{
    return RBOOL(Jed_Backup_By_Copying);
}

static void backup_by_copy_setter(VALUE val)
{
    Jed_Backup_By_Copying = FIX2INT(val);
}
#endif /* REAL_UNIX_SYSTEM */

static VALUE minibuf_active_getter()
{
    return RBOOL(MiniBuffer_Active);
}

static VALUE display_8bit_getter()
{
    return RBOOL(Display_Eight_Bit);
}

static void display_8bit_setter(VALUE val)
{
    Display_Eight_Bit = val;
}

#ifndef Sixteen_Bit_System
static VALUE kill_array_size_getter()
{
    return INT2FIX(Kill_Array_Size);
}

static void kill_array_size_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Kill_Array_Size = FIX2INT(val);
}
#endif /* Sixteen_Bit_System */

static VALUE blink_getter()
{
    return RBOOL(Blink_Flag);
}

static void blink_setter(VALUE val)
{
    Blink_Flag = val;
}

static VALUE wrap_indents_getter()
{
    return RBOOL(Indented_Text_Mode);
}

static void wrap_indents_setter(VALUE val)
{
    Indented_Text_Mode = val;
}

static VALUE tab_default_getter()
{
    return INT2FIX(Jed_Tab_Default);
}

static void tab_default_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Jed_Tab_Default = FIX2INT(val);
}

static VALUE ignore_beep_getter()
{
    return INT2FIX(*tt_Ignore_Beep);
}

static void ignore_beep_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    *tt_Ignore_Beep = FIX2INT(val);
}

static VALUE add_newline_getter()
{
    return RBOOL(Require_Final_Newline);
}

static void add_newline_setter(VALUE val)
{
    Require_Final_Newline = val;
}

static VALUE display_time_getter()
{
    return RBOOL(Display_Time);
}

static void display_time_setter(VALUE val)
{
    Display_Time = val;
}

static VALUE want_eob_getter()
{
    return RBOOL(Want_Eob);
}

static void want_eob_setter(VALUE val)
{
    Want_Eob = val;
}

static VALUE wrap_getter()
{
    return INT2FIX(Jed_Wrap_Column);
}

static void wrap_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Jed_Wrap_Column = FIX2INT(val);
}

static VALUE meta_char_getter()
{
    return INT2FIX(Meta_Char);
}

static void meta_char_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Meta_Char = FIX2INT(val);
}

static VALUE dec_8bit_hack_getter()
{
    return INT2FIX(DEC_8Bit_Hack);
}

static void dec_8bit_hack_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    DEC_8Bit_Hack = FIX2INT(val);
}

static VALUE batch_getter()
{
    return RBOOL(Batch);
}

static VALUE last_char_getter()
{
    return INT2FIX(SLang_Last_Key_Char);
}

static VALUE max_hits_getter()
{
    return INT2FIX(Jed_Max_Hits);
}

static VALUE point_getter()
{
    return INT2FIX(Point);
}

static void point_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Point = FIX2INT(val);
}

static VALUE ignore_user_abort_getter()
{
    return RBOOL(Ignore_User_Abort);
}

static void ignore_user_abort_setter(VALUE val)
{
    Ignore_User_Abort = val;
}

static VALUE kill_line_feature_getter()
{
    return RBOOL(Kill_Line_Feature);
}

static void kill_line_feature_setter(VALUE val)
{
    Kill_Line_Feature = val;
}

#ifdef pc_system
static VALUE alt_char_getter()
{
    return INT2FIX(PC_Alt_Char);
}

static void alt_char_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    PC_Alt_Char = FIX2INT(val);
}
#endif /* pc_system */

static VALUE highlight_getter()
{
    return RBOOL(Wants_Attributes);
}

static void highlight_setter(VALUE val)
{
    Wants_Attributes = val;
}

static VALUE horizontal_pan_getter()
{
    return RBOOL(Wants_HScroll);
}

static void horizontal_pan_setter(VALUE val)
{
    Wants_HScroll = val;
}

static VALUE want_syntax_highlight_getter()
{
    return RBOOL(Wants_Syntax_Highlight);
}

static void want_syntax_highlight_setter(VALUE val)
{
    Wants_Syntax_Highlight = val;
}

static VALUE dollar_char_getter()
{
    return INT2FIX(Jed_Dollar);
}

static void dollar_char_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    Jed_Dollar = FIX2INT(val);
}

static VALUE jed_codedetect_getter()
{
    return INT2FIX(file_kanjicode_detect);
}

static void jed_codedetect_setter(VALUE val)
{
    Check_Type(val, T_FIXNUM);
    file_kanjicode_detect = FIX2INT(val);
}

static VALUE replace_preserve_case_getter()
{
    return RBOOL(Replace_Preserve_Case);
}

static void replace_preserve_case_setter(VALUE val)
{
    Replace_Preserve_Case = val;
}

static VALUE case_search_getter()
{
    return RBOOL(Case_Sensitive);
}

static void case_search_setter(VALUE val)
{
    Case_Sensitive = val;
}

static VALUE use_tabs_getter()
{
    return RBOOL(Jed_Use_Tabs);
}

static void use_tabs_setter(VALUE val)
{
    Jed_Use_Tabs = val;
}

static VALUE defining_marcro_getter()
{
    return RBOOL(Defining_Keyboard_Macro);
}

static void defining_marcro_setter(VALUE val)
{
    Defining_Keyboard_Macro = val;
}

static VALUE executing_macro_getter()
{
    return RBOOL(Executing_Keyboard_Macro);
}

static void executing_macro_setter(VALUE val)
{
    Executing_Keyboard_Macro = val;
}

void init_ruby_lib()
{
    rb_undef_method(rb_mKernel, "exit");

    mJed = rb_define_module("Jed");
    rb_define_module_function(mJed, "evaluate_cmd", mjed_evaluate_cmd, 0);
    rb_define_module_function(mJed, "internal_function?",
			      mjed_internal_function_p, 1);
    rb_define_module_function(mJed, "call_internal_function",
			      mjed_call_internal_function, 1);
    rb_define_module_function(mJed, "defined_in_slang?",
			      mjed_defined_in_slang_p, 1);
    rb_define_module_function(mJed, "eval_slang_string",
			      mjed_eval_slang_string, 1);
    rb_define_module_function(mJed, "exit_jed", mjed_exit_jed, 0);
    rb_define_module_function(mJed, "quit_jed", mjed_quit_jed, 0);
    rb_define_module_function(mJed, "message", mjed_message, -1);
    rb_define_module_function(mJed, "flush_message",
			      mjed_flush_message, -1);
    rb_define_module_function(mJed, "set_status_line",
			      mjed_set_status_line, 2);
    rb_define_module_function(mJed, "update", mjed_update, 1);
    rb_define_module_function(mJed, "show_error", mjed_show_error, -1);
    rb_define_module_function(mJed, "read_from_minibuffer",
			      mjed_read_from_minibuffer, 3);
    rb_define_module_function(mJed, "read_with_completion",
			      mjed_read_with_completion, 4);
    rb_define_module_function(mJed, "beep", mjed_beep, 0);
    rb_define_module_function(mJed, "input_pending",
			      mjed_input_pending, 1);
    rb_define_module_function(mJed, "send_string_to_terminal",
			      mjed_send_string_to_terminal, 1);
    rb_define_module_function(mJed, "pause", mjed_pause, 1);
    rb_define_module_function(mJed, "recenter", mjed_recenter, -1);
    rb_define_module_function(mJed, "get_yes_no", mjed_get_yes_no, 1);
    rb_define_module_function(mJed, "find_file", mjed_find_file, -1);
    rb_define_module_function(mJed, "create_keymap", mjed_create_keymap, 1);
    rb_define_module_function(mJed, "copy_keymap", mjed_copy_keymap, 2);
    rb_define_module_function(mJed, "define_key", mjed_define_key, -1);
    rb_define_module_function(mJed, "undefine_key", mjed_undefine_key, -1);
    rb_define_module_function(mJed, "getkey", mjed_getkey, 0);
    rb_define_module_function(mJed, "ungetkey", mjed_ungetkey, 1);
    rb_define_module_function(mJed, "add_completion",
			      mjed_add_completion, 1);
    rb_define_module_function(mJed, "set_color", mjed_set_color, 3);
    rb_define_module_function(mJed, "set_mode", mjed_set_mode, 2);
    rb_define_const(mJed, "NO_MODE", INT2FIX(0));
    rb_define_const(mJed, "WRAP_MODE", INT2FIX(1));
    rb_define_const(mJed, "C_MODE", INT2FIX(2));
    rb_define_const(mJed, "LANGUAGE_MODE", INT2FIX(4));
    rb_define_const(mJed, "SLANG_MODE", INT2FIX(8));
    rb_define_const(mJed, "FORTRAN_MODE", INT2FIX(16));
    rb_define_const(mJed, "TEX_MODE", INT2FIX(32));
    rb_define_module_function(mJed, "keymap?", mjed_keymap_p, 1);
    rb_define_module_function(mJed, "current_keymap", mjed_what_keymap, 0);
    rb_define_module_function(mJed, "use_keymap", mjed_use_keymap, 1);
    rb_define_module_function(mJed, "dump_key_bindings",
			      mjed_dump_key_bindings, 1);
    rb_define_module_function(mJed, "create_syntax_table",
			      mjed_create_syntax_table, 1);
    rb_define_module_function(mJed, "define_syntax",
			      mjed_define_syntax, -1);
    rb_define_module_function(mJed, "use_syntax_table",
			      mjed_use_syntax_table, 1);
#if JED_HAS_DFA_SYNTAX
    rb_define_module_function(mJed, "define_highlight_rule",
			      mjed_define_highlight_rule, 3);
    rb_define_module_function(mJed, "build_highlight_table",
			      mjed_build_highlight_table, 1);
    rb_define_module_function(mJed, "enable_highlight_cache",
			      mjed_enable_highlight_cache, 2);
#endif /* JED_HAS_DFA_SYNTAX */
    rb_define_module_function(mJed, "define_keywords_n",
			      mjed_define_keywords, 4);
    rb_define_module_function(mJed, "add_hook", mjed_add_hook, -1);
    rb_define_module_function(mJed, "run_hooks", mjed_run_hooks, -1);
    rb_include_module(rb_cObject, mJed);

#if !defined(pc_system) || defined(__GO32__)
    rb_define_virtual_variable("$USE_ANSI_COLORS",
			       use_ansi_colors_getter,
			       use_ansi_colors_setter);
#endif /* !defined(pc_system) || defined(__GO32__) */
#ifdef REAL_UNIX_SYSTEM
    rb_define_virtual_variable("$BACKUP_BY_COPYING",
			       backup_by_copy_getter,
			       backup_by_copy_setter);
#endif /* REAL_UNIX_SYSTEM */
    rb_define_virtual_variable("$MINIBUFFER_ACTIVE",
			       minibuf_active_getter, 0);
    rb_define_virtual_variable("$DISPLAY_8BIT",
			       display_8bit_getter,
			       display_8bit_setter);
#ifndef Sixteen_Bit_System
    rb_define_virtual_variable("$KILL_ARRAY_SIZE",
			       kill_array_size_getter,
			       kill_array_size_setter);
#endif /* Sixteen_Bit_System */
    rb_define_virtual_variable("$BLINK",
			       blink_getter,
			       blink_setter);
    rb_define_virtual_variable("$WRAP_INDENTS",
			       wrap_indents_getter,
			       wrap_indents_setter);
    rb_define_virtual_variable("$TAB_DEFAULT",
			       tab_default_getter,
			       tab_default_setter);
    rb_define_virtual_variable("$IGNORE_BEEP",
			       ignore_beep_getter,
			       ignore_beep_setter);
    rb_define_virtual_variable("$ADD_NEWLINE",
			       add_newline_getter,
			       add_newline_setter);
    rb_define_virtual_variable("$DISPLAY_TIME",
			       display_time_getter,
			       display_time_setter);
    rb_define_virtual_variable("$WANT_EOB",
			       want_eob_getter,
			       want_eob_setter);
    rb_define_virtual_variable("$WRAP",
			       wrap_getter,
			       wrap_setter);
    rb_define_virtual_variable("$META_CHAR",
			       meta_char_getter,
			       meta_char_setter);
    rb_define_virtual_variable("$DEC_8bit_hack",
			       dec_8bit_hack_getter,
			       dec_8bit_hack_setter);
    rb_define_virtual_variable("$BATCH",
			       batch_getter, 0);
    rb_define_virtual_variable("$LAST_CHAR",
			       last_char_getter, 0);
    rb_define_virtual_variable("$MAX_HITS",
			       max_hits_getter, 0);
    rb_define_virtual_variable("$POINT",
			       point_getter,
			       point_setter);
    rb_define_virtual_variable("$IGNORE_USER_ABORT",
			       ignore_user_abort_getter,
			       ignore_user_abort_setter);
    rb_define_virtual_variable("$KILL_LINE_FEATURE",
			       kill_line_feature_getter,
			       kill_line_feature_setter);
#ifdef pc_system
    rb_define_virtual_variable("$ALT_CHAR",
			       alt_char_getter,
			       alt_char_setter);
#endif
    rb_define_virtual_variable("$HIGHLIGHT",
			       highlight_getter,
			       highlight_setter);
    rb_define_virtual_variable("$HORIZONTAL_PAN",
			       horizontal_pan_getter,
			       horizontal_pan_setter);
    rb_define_virtual_variable("$WANT_SYNTAX_HIGHLIGHT",
			       want_syntax_highlight_getter,
			       want_syntax_highlight_setter);
    rb_define_virtual_variable("$DOLLAR_CHAR",
			       dollar_char_getter,
			       dollar_char_setter);
    rb_define_virtual_variable("$JED_CODEDETECT",
			       jed_codedetect_getter,
			       jed_codedetect_setter);
    rb_define_virtual_variable("$REPLACE_PRESERVE_CASE",
			       replace_preserve_case_getter,
			       replace_preserve_case_setter);
    rb_define_virtual_variable("$CASE_SEARCH",
			       case_search_getter,
			       case_search_setter);
    rb_define_virtual_variable("$USE_TABS",
			       use_tabs_getter,
			       use_tabs_setter);
    rb_define_virtual_variable("$DEFINING_MACRO",
			       defining_marcro_getter,
			       defining_marcro_setter);
    rb_define_virtual_variable("$EXECUTING_MACRO",
			       executing_macro_getter,
			       executing_macro_setter);

    rb_define_const(mJed, "JED_ROOT", rb_str_new2(Jed_Root_Dir));
    rb_define_const(mJed, "JED_VERSION", INT2FIX(Jed_Version_Number));
    rb_define_const(mJed, "JED_JVERSION", INT2FIX(Jed_JVersion_Number));
    rb_define_const(mJed, "JED_RVERSION", INT2FIX(Jed_RVersion_Number));
#if JED_HAS_DFA_SYNTAX
    rb_define_const(mJed, "HAS_DFA_SYNTAX", Qtrue);
#endif /* JED_HAS_DFA_SYNTAX */

    rb_hook_list = rb_hash_new();
    rb_global_variable(&rb_hook_list);

    Init_rb_buffer();
    Init_rb_window();
#ifdef JED_HAS_ABBREVS
    Init_rb_abbrev();
#endif /* JED_HAS_ABBREVS */
}
