/* Copyright (c) 1992, 1998 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
#include "config.h"
#include "jed-feat.h"

#ifdef JED_HAS_SUBPROCESSES

# include <stdio.h>
# include <slang.h>

# include "jdmacros.h"

# include <stdlib.h>

# include "jprocess.h"
# include "buffer.h"
# include "ins.h"
# include "ledit.h"
# include "misc.h"
# include "paste.h"
# include "kanji.c"

int Num_Subprocesses;
CRITICAL_SECTION Critical_Section;
#if WIN95BUG
HANDLE Input_Events[MAX_PROCESSES * 2 + 1];
#else
HANDLE Input_Events[MAX_PROCESSES + 1];
#endif

int kanji_default_process_code = KANJI_DEFAULT_CODE;

static int Subprocess_Id[MAX_PROCESSES + 1];

volatile int Child_Status_Changed_Flag;/* if this is non-zero, editor
					* should call the appropriate
					* function below to call slang
					* handlers.
					*/

typedef struct
{
   int flags;			       /* This is zero if the process is gone
					* and the status is nolonger avail */
# define PROCESS_RUNNING		1
# define PROCESS_STOPPED		2
# define PROCESS_ALIVE		3
# define PROCESS_EXITED		4
# define PROCESS_SIGNALLED	8
   int return_status;		       /* This value depends on the flags */

   int status_changed;		       /* non-zero if status changed. */
   HANDLE rd, wd;		       /* read/write handles */
   int rd_kanji_code, wd_kanji_code;   /* read/write kanji code */
   HANDLE hprocess;		       /* real process handle */
   
#if WIN95BUG
   HANDLE hthread;
#endif
   
   int output_type;
# define PROCESS_USE_BUFFER	1
# define PROCESS_USE_SLANG	2
# define PROCESS_SAVE_POINT	4
# define PROCESS_AT_POINT	8
   Buffer *buffer;		       /* buffer associated with process */
   SLang_Name_Type *slang_fun;	       /* function to pass output to */
   SLang_MMT_Type *umark;	       /* marks point of last output */

   SLang_Name_Type *status_change_fun; /* call this if process status changes
					* The function should be declared like
					* define fun (pid, flags);
					* The flags parameter corresponds to
					* the flags field in this struct and
					* the pid is NOT the pid of this struct
					*/
   HANDLE input_event;
# define PROCESS_BUFSIZE 513
   unsigned char input_buf [PROCESS_BUFSIZE];
   int input_bufsize;
}
Process_Type;

static Process_Type Processes[MAX_PROCESSES];

static Process_Type *get_process (int fd)
{
   Process_Type *p;

   if ((fd >= 0) && (fd < MAX_PROCESSES)
       && (p = &Processes[fd], p->flags != 0)) return p;

   msg_error ("process does not exist.");
   return NULL;
}

static void call_slang_status_change_hook (Process_Type *p)
{
   Buffer *cbuf = CBuf;
   if ((p->status_change_fun == NULL) || (p->buffer == NULL)) return;

   cbuf->locked++;
   switch_to_buffer (p->buffer);
   SLang_push_integer ((int) (p - Processes));
   SLang_push_integer (p->flags);
   SLexecute_function (p->status_change_fun);
   touch_screen ();
   if (CBuf != cbuf) switch_to_buffer (cbuf);
   cbuf->locked--;
}

static void close_rd_and_wd(Process_Type *p)
{
   char eof = '\x1a';		       /* ^Z */
   DWORD d;

   if (p->wd != INVALID_HANDLE_VALUE)
     {
	WriteFile(p->wd, &eof, 1, &d, NULL);
	CloseHandle(p->wd);
	p->wd = INVALID_HANDLE_VALUE;
     }

   if (p->rd != INVALID_HANDLE_VALUE)
     {
	CloseHandle(p->rd);
	p->rd = INVALID_HANDLE_VALUE;
     }
}

/* This routine is called to clean up after the process has exited.
 * After getting the exit status, we call a slang hook and if the
 * process is dead, adjust the process arrays to delete the process.
 */

static void get_process_status (Process_Type *p)
{
   int i;
   
#if WIN95BUG
   int n;
#endif
   
   /* Call slang to let it know what happened.  Do it first before we
    * really shut it down to give the hook a chance to query the state of
    * it before it returns.
    */
   call_slang_status_change_hook (p);
   if (p->flags & PROCESS_ALIVE) return;

   /* Process is dead.  So perform clean up. */
   CloseHandle(p->input_event);

   if (p->buffer != NULL) p->buffer->subprocess = 0;

   if (p->umark != NULL) SLang_free_mmt (p->umark);

   /* Adjust the array of input events. Note that first handle
    * is used for read descriptor for console jed */
   i = 1;
   while (i <= Num_Subprocesses)
     {
	if (Input_Events[i] == p->input_event) break;
	i++;
     }
   
   Num_Subprocesses--;
#if WIN95BUG   
   n = i + Num_Subprocesses;
#endif   
   
   while (i <= Num_Subprocesses)
     {
	Input_Events[i] = Input_Events[i + 1];
	Subprocess_Id[i] = Subprocess_Id[i + 1];
	i++;
     }
   
#if WIN95BUG
   i = n;
   while (i <= Num_Subprocesses * 2)
     {
	Input_Events[i] = Input_Events[i + 2];
	i++;
     }
#endif

   memset((char *) p, 0, sizeof(Process_Type));
}

int jed_close_process (int *fd)
{
   Process_Type *p;
   char *dstbuf;
   int n;

   if (NULL == (p = get_process (*fd))) return -1;

   EnterCriticalSection(&Critical_Section);
   TerminateProcess(p->hprocess, 0);
   if (p->buffer != NULL) p->buffer->subprocess = 0;
   LeaveCriticalSection(&Critical_Section);

   return 0;
}

void jed_kill_process (int fd)
{
   /* This function is called when the buffer is going to be destroyed */
   Processes[fd].buffer = NULL;
   jed_close_process (&fd);
}

void jed_get_child_status (void)
{
   Process_Type *p, *pmax;

   p = Processes; pmax = p + MAX_PROCESSES;

   while (p < pmax)
     {
	if (p->flags && p->status_changed)
	  {
	     EnterCriticalSection(&Critical_Section);
	     Child_Status_Changed_Flag--;
	     LeaveCriticalSection(&Critical_Section);
	     p->status_changed--;
	     get_process_status (p);
	  }
	p++;
     }
}

static DWORD thread_func(int fd)
{
   char buf[513];
   DWORD n;
   int i;
   Process_Type *p = get_process(fd);
 
   while (1)
     {
	EnterCriticalSection(&Critical_Section);
	i = PROCESS_BUFSIZE - p->input_bufsize - 1;
	LeaveCriticalSection(&Critical_Section);
	
	/* Check if there is free space in process buffer */
	if (i)
	  {
	     if (!ReadFile(p->rd, buf, i, &n, NULL))
	       {
		  /* read handle is closed, so mark process as terminated */
		  p->rd = INVALID_HANDLE_VALUE;
		  p->flags = PROCESS_EXITED;
		  p->status_changed++;
		  SetEvent(p->input_event);
		  return 0;
	       }
	     
	     EnterCriticalSection(&Critical_Section);
	     buf[n] = 0;
	     memcpy(p->input_buf + p->input_bufsize, buf, n + 1);
	     p->input_bufsize += n;
	     SetEvent(p->input_event);
	     LeaveCriticalSection(&Critical_Section);
	  }
     }
   
   /* should never be reached */
   return 0;
}

void read_process_input(int fd)
{
   Buffer *b = CBuf, *pbuf;
   int otype, n, status;
   Process_Type *p;
   char buf[PROCESS_BUFSIZE];
   
   /* Should never happen */
   if (NULL == (p = get_process (Subprocess_Id[fd]))) return;
   
#if WIN95BUG
   if ((p->rd != INVALID_HANDLE_VALUE) && 
       (WaitForSingleObject(p->hprocess, 0) != WAIT_TIMEOUT))
     {
	EnterCriticalSection(&Critical_Section);
	p->status_changed++;
	p->flags = PROCESS_EXITED;
	CloseHandle(p->rd);
	p->rd = INVALID_HANDLE_VALUE;
	LeaveCriticalSection(&Critical_Section);
	TerminateThread(p->hthread, 0);
     }
#endif   
   
   EnterCriticalSection(&Critical_Section);
   ResetEvent(p->input_event);
   n = p->input_bufsize;
   memcpy(buf, p->input_buf, n + 1);
   p->input_bufsize = 0;
   *p->input_buf = 0;
   status = p->status_changed;
   LeaveCriticalSection(&Critical_Section);
   
   if (n > 0)
     {
	otype = p->output_type;
	pbuf = p->buffer;
	
	if (pbuf != NULL)
	  {
	     switch_to_buffer (pbuf);
	     pbuf->locked++;
	  }
	
	if (otype & PROCESS_SAVE_POINT) push_spot ();
	
	if (otype & PROCESS_USE_BUFFER)
	  {
	     if (0 == (otype & PROCESS_AT_POINT)) eob ();
	     dstbuf = KanjiCodeConv(buf, &n, p->rd_kanji_code,
				    kSLcode, SKanaToDKana);
	     ins_chars(dstbuf, n);
	     if(dstbuf != buf) SLfree(dstbuf);
	     
	     jed_move_user_object_mark (p->umark);
	  }
	else if (otype == PROCESS_USE_SLANG)
	  {
	     SLang_push_integer ((int) (p - Processes));
	     SLang_push_string ((char *) buf);
	     SLexecute_function (p->slang_fun);    /* function to pass output to */
	  }
	
	if (otype & PROCESS_SAVE_POINT) pop_spot ();
	else if (otype & PROCESS_USE_BUFFER) move_window_marks (0);
	
	if (p->buffer != NULL)
	  {
	     if (b != CBuf) switch_to_buffer (b);
	     pbuf->locked--;
	  }
     }
   
   if (status)
     {
	p->status_changed--;
	get_process_status (p);
     }
   touch_screen ();
}

static int open_process (char *pgm, char **argv)
{
   int pd;
   HANDLE fds0[2], fds1[2];
   Process_Type *p;
   char **arg, *cmd_line, *cp;
   SLang_MMT_Type *mmt;
   STARTUPINFO si;
   PROCESS_INFORMATION pi;
   SECURITY_ATTRIBUTES sa;
   DWORD d, id_thread;

#if WIN95BUG
   int i;
#endif   
   
   pd = 0; while ((pd < MAX_PROCESSES) && Processes[pd].flags) pd++;
   if (pd == MAX_PROCESSES) return -1;
   p = &Processes[pd];
   
   SLMEMSET ((char *) p, 0, sizeof (Process_Type));
   
   sa.nLength = sizeof(SECURITY_ATTRIBUTES);
   sa.bInheritHandle = Qtrue;
   sa.lpSecurityDescriptor = NULL;
   
   if (CreatePipe(&fds0[0], &fds0[1], &sa, 0))
     {
	if (!CreatePipe(&fds1[0], &fds1[1], &sa, 0))
	  {
	     CloseHandle(fds0[0]);
	     CloseHandle(fds0[1]);
	     return -1;
	  }
     }
   else
     return -1;
   
   if (NULL == (mmt = jed_make_user_object_mark ())) return -1;
   
   si.cb = sizeof(STARTUPINFO);
   si.lpReserved = NULL;
   si.lpReserved2 = NULL;
   si.cbReserved2 = 0;
   si.lpDesktop = NULL;
   si.lpTitle = NULL;
   si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
   si.wShowWindow = SW_MINIMIZE;
   si.hStdInput = fds0[0];
   si.hStdOutput = fds1[1];
   si.hStdError = fds1[1];
   
   d = strlen(pgm) + 1;
   arg = argv;
   while (*arg != NULL)
     {
	d += strlen(*arg) + 1;
	arg++;
     }
   
   if (NULL == (cmd_line = SLmalloc(d)))
     {
	SLang_free_mmt (mmt);
	return -1;
     }
   
   d = strlen (pgm);
   cp = cmd_line;
   
   arg = argv;
   while (*arg != NULL)
     {
	d = strlen(*arg);
	SLMEMCPY(cp, *arg, d);
	cp += d;
	*cp++ = ' ';
	arg++;
     }
   
   *cp = 0;
   
   if (CreateProcess(NULL, cmd_line, NULL, NULL, Qtrue, 
		     CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE, NULL, NULL, &si, &pi))
     {
	CloseHandle(fds0[0]);
	CloseHandle(fds1[1]);
	
	p->flags = PROCESS_RUNNING;
	p->rd = fds1[0];
	p->wd = fds0[1];
	p->hprocess = pi.hProcess;
	p->rd_kanji_code = p->wd_kanji_code = kanji_default_process_code;
	
#if WIN95BUG
	i = Num_Subprocesses * 2;
	while (i > Num_Subprocesses)
	  {
	     Input_Events[i + 1] = Input_Events[i];
	     i--;
	  }
#endif
	
	Num_Subprocesses += 1;
	
	CBuf->subprocess = pd + 1;
	
	/* Processing options */
	p->buffer = CBuf;
	p->output_type = PROCESS_USE_BUFFER;
	p->umark = mmt;
	p->input_bufsize = 0;
	SLang_inc_mmt (mmt);	       /* tell slang we are keeping a copy */
	
	p->input_event = CreateEvent(NULL, Qtrue, Qfalse, NULL);
	Input_Events[Num_Subprocesses] = p->input_event;
	
#if WIN95BUG
	Input_Events[Num_Subprocesses * 2] = p->hprocess;
#endif
	
	Subprocess_Id[Num_Subprocesses] = pd;
	
#if WIN95BUG	
	p->hthread =
#endif	  
	  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)thread_func, (LPVOID) pd, 0, &id_thread);
     }
   else
     {
	jed_verror ("open_process: exec of %s failed.", cmd_line);
	CloseHandle(fds0[0]);
	CloseHandle(fds0[1]);
	CloseHandle(fds1[0]);
	CloseHandle(fds1[1]);
	
	SLang_free_mmt(mmt);
	return -1;
     }
   
   return pd;
}

int jed_send_process (int *fd, char *str)
{
   DWORD d;
   char *dststr;
   int n;
   
   Process_Type *p = get_process (*fd);
   if ((p == NULL) || (p->wd == INVALID_HANDLE_VALUE)) return -1;
   n = strlen(str);
   dststr = KanjiCodeConv(str, &n, kSLcode, p->wd_kanji_code, SKanaToDKana);
       WriteFile(p->wd, dststr, n, &d, NULL);
   if(dststr != str)
     SLfree(dststr);
   return 0;
}

void jed_send_process_eof (int *fd)
{
   DWORD d;
   
   Process_Type *p = get_process (*fd);
   if (p == NULL) return;
   
   if (p->wd != INVALID_HANDLE_VALUE)
     {
	CloseHandle(p->wd);
	p->wd = INVALID_HANDLE_VALUE;
     }
}

void jed_set_process (int *pd, char *what, char *s)
{
   Process_Type *p;
   SLang_Name_Type *f;
   
   if (NULL == (p = get_process (*pd))) return;
   if (!strcmp (what, "output"))
     {
	if (*s == '.') p->output_type = PROCESS_AT_POINT | PROCESS_USE_BUFFER;
	else if (*s == '@') p->output_type = PROCESS_SAVE_POINT | PROCESS_USE_BUFFER;
	else if (*s && (NULL != (f = SLang_get_function (s))))
	  {
	     p->output_type = PROCESS_USE_SLANG;
	     p->slang_fun = f;
	  }
	else p->output_type = PROCESS_USE_BUFFER;
     }
   else if (!strcmp (what, "signal"))
     {
	if (*s && (NULL != (f = SLang_get_function (s))))
	  {
	     p->status_change_fun = f;
	  }
     }
}

void jed_get_process_mark (int *fd)
{
   Process_Type *p;
   if (NULL == (p = get_process (*fd))) return;
   
   SLang_push_mmt (p->umark);
}

int jed_open_process (int *np)
{
   int fd = -1;
   char *argv[502];
   int n = *np;
   
   if (CBuf->subprocess)
     {
	msg_error ("There is already a process attached to this buffer.");
	return -1;
     }
   
   if ((n > 500) || (n < 0))
     {
	msg_error ("Arguments out of range.");
	return -1;
     }
   
   n++;				       /* for argv0 since *np does not include
					* it.
					*/
   argv[n] = NULL;
   while (n--)
     {
	if (SLpop_string (&argv[n]))
	  {
	     n++;
	     goto free_return;
	  }
     }
   n = 0;
   
   if ((fd = open_process(argv[0], argv)) < 0)
     {
	msg_error ("Unable to open process.");
     }
   
   /* free up the argument strings */
   free_return:
   
   while (n <= *np)
     {
	SLfree (argv[n]);
	n++;
     }
   
   return fd;
}

int jed_set_kanji_process_code (int *fd, int *rw, int *code) /*{{{*/
				/* rw: set read/write process code
				 *   1: read
				 *   2: write
				 *   3: read and write
				 * code: KANJI code (ASCII/EUC/JIS/SJIS)
				 */
{
   Process_Type *p;
   int kanji_code = *code;
   
   if (NULL == (p = get_process (*fd))) return -1;
   if(kanji_code < ASCII || SJIS < kanji_code) return -1;
   if(*rw & 1) p->rd_kanji_code = kanji_code;
   if(*rw & 2) p->wd_kanji_code = kanji_code;
   
   return 0;
}
#endif
int jed_signal_process (int *fd, int *sig) /*{{{*/
{
   return -1;
}

#if defined (__BORLANDC__) || defined (_MSC_VER)
# define popen msw_popen
# define pclose msw_pclose
extern FILE *msw_popen(char *, char *);
extern int msw_pclose(FILE *);
#endif

FILE *jed_popen (char *cmd, char *type) /*{{{*/
{
   return popen(cmd, type);
}

/*}}}*/

int jed_pclose (FILE *fp) /*{{{*/
{
   return pclose (fp);
}

#endif
